# <@LICENSE>
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to you under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at:
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </@LICENSE>

=head1 NAME

RWL - recipient white list plugin

=head1 SYNOPSIS

  loadplugin     Mail::SpamAssassin::Plugin::RWL
  header         KNOWN_ADDRESS eval:check_rwl_plugin()
  score		 KNOWN_ADDRESS   -3.0
=head1 DESCRIPTION

To try this plugin, write the above three lines in the synopsis to
C</etc/mail/spamassassin/init.pre>.

=cut

package Mail::SpamAssassin::Plugin::RWL;

use Mail::SpamAssassin::Plugin;
use Mail::SpamAssassin::Logger;
use DBI;
use DBD::Pg;
use strict;
use warnings;
use bytes;

use vars qw(@ISA);
@ISA = qw(Mail::SpamAssassin::Plugin);

# constructor: register the eval rule
sub new {
  my $class = shift;
  my $mailsaobject = shift;
# logger for debugging
#  Mail::SpamAssassin::Logger::add(method => 'file', filename => '/var/log/spamassassin.log');

  # some boilerplate...
  $class = ref($class) || $class;
  my $self = $class->SUPER::new($mailsaobject);
  bless ($self, $class);
  # the important bit!
  $self->register_eval_rule ("check_rwl_plugin");
  #read config
  $self->set_config($mailsaobject->{conf});
  Mail::SpamAssassin::Logger::info ("registered Mail::SpamAssassin::Plugin::RWL: $self\n");
  $self->{RWL_dbh} = undef;
  $self->{Hour} = -1;
  return $self;
}

#destructor
sub finish {
  my $self = shift;
  $self->{RWL_dbh}->disconnect if ($self->{RWL_dbh});
  %{$self} = ();
Mail::SpamAssassin::Logger::close_log();
}

sub set_config {
  my($self, $conf) = @_;
    my @cmds = ();
    
=head1 USER OPTIONS
    
The following configuration settings are used to control auto-learning:
    
=over 4
    
=item rwl_sql_dsn string  (default: dbi:Pg:dbname=database;host=localhost;port=5432)

Connection string to database with list of local mail domains and users

=cut
    
  push (@cmds, {
      setting => 'rwl_sql_dsn',
      default => "dbi:Pg:dbname=database;host=localhost;port=5432",
      type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
    });
		    
=item rwl_sql_username string    (default: username)

User name for create connection to database 		    

=cut
		    
  push (@cmds, {
      setting => 'rwl_sql_username',
      default => "username",
      type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
    });

=item rwl_sql_password string    (default: password)

User password for create connection to database 	    

=cut
  push (@cmds, {
      setting => 'rwl_sql_password',
      default => "password",
      type => $Mail::SpamAssassin::Conf::CONF_TYPE_STRING
    });


 $conf->{parser}->register_commands(\@cmds);
 }

sub reconnect_to_db
  {
  #create db connection
  my ($self) = @_;
  my $db_user  = $self->{main}->{conf}->{rwl_sql_username};
  my $db_pass  = $self->{main}->{conf}->{rwl_sql_password};
  my $sql_dsn  = $self->{main}->{conf}->{rwl_sql_dsn};
  my $rv;
  my $dbh;
  Mail::SpamAssassin::Logger::info ("RWL: SQL state: $DBI::errstr \n") if($DBI::errstr);
  $self->{RWL_dbh}->disconnect if ($self->{RWL_dbh});
  $dbh = DBI->connect("$sql_dsn", "$db_user", "$db_pass");
  if(!$dbh)
    {
    Mail::SpamAssassin::Logger::log_message ("error", "RWL: error: $DBI::errstr \n");
    die "Oops!!!";
    }
  $self->{RWL_dbh} = $dbh;
  Mail::SpamAssassin::Logger::info ("RWL: reconnect to db\n");
  }

sub connect_to_db
  {
  #create db connection
  my ($self) = @_;
  my $db_user  = $self->{main}->{conf}->{rwl_sql_username};
  my $db_pass  = $self->{main}->{conf}->{rwl_sql_password};
  my $sql_dsn  = $self->{main}->{conf}->{rwl_sql_dsn};
  my $rv;
  my $dbh;
  $dbh = DBI->connect("$sql_dsn", "$db_user", "$db_pass");
  if(!$dbh)
    {
    Mail::SpamAssassin::Logger::log_message ("error", "RWL: error: $DBI::errstr \n");
    die "Oops!!!";
    }
  $self->{RWL_dbh} = $dbh;
  Mail::SpamAssassin::Logger::info ("RWL: connect to db\n");

  $rv=$dbh->do("
CREATE TABLE IF NOT EXIST list
(
  email character varying(128) NOT NULL,
  counter integer,
  first_detect timestamp without time zone,
  last_detect timestamp without time zone,
  do_not_delete boolean DEFAULT false,
  CONSTRAINT list_pkey PRIMARY KEY (email)
)
WITH (OIDS=FALSE);
  ");
	 Mail::SpamAssassin::Logger::info ("RWL: SQL state: $DBI::errstr \n") if($DBI::errstr);

  $rv=$dbh->do("
CREATE TABLE IF NOT EXIST domains
(
  name character varying(128) NOT NULL,
  CONSTRAINT domains_pkey PRIMARY KEY (name)
)
WITH (OIDS=FALSE);
  ");
	  Mail::SpamAssassin::Logger::info ("RWL: SQL state: $DBI::errstr \n") if($DBI::errstr);

  $rv=$dbh->do("
CREATE OR REPLACE FUNCTION list_update(text, timestamp)
  RETURNS integer AS
  \$BODY\$
  BEGIN
    UPDATE list SET counter = counter + 1, last_detect = \$2 WHERE email = lower(\$1); 
    IF FOUND = FALSE THEN
      INSERT INTO list (email, counter, first_detect, last_detect) VALUES (lower(\$1), 1, \$2, \$2);
      RETURN 1;
    END IF;
    RETURN 2;
  END;
  \$BODY\$
LANGUAGE 'plpgsql' VOLATILE;
  ");
	  Mail::SpamAssassin::Logger::info ("RWL: SQL state: $DBI::errstr \n") if($DBI::errstr);

  Mail::SpamAssassin::Logger::info ("RWL: SQL init complete\n");
  }

sub cron_tasks
  {
  my ($self) = @_;
  my $rv;

# db cleaning
  Mail::SpamAssassin::Logger::info ("RWL: Deleting outdated entries\n");
#  $rv=$self->{RWL_dbh}->prepare("SELECT * FROM list WHERE \'now\'::timestamp > last_detect + \'3 month\'::interval * (1 + log(counter))");
  $rv=$self->{RWL_dbh}->prepare("DELETE FROM list WHERE \'now\'::timestamp > last_detect + \'3 month\'::interval * (1 + log(counter))");
  $rv->execute;
  my $result = $rv->rows;
  Mail::SpamAssassin::Logger::info ("RWL: $result entries deleted\n");
  $rv->finish() if($rv);
  reconnect_to_db($self) if($result < 0);

# read domains from SQL into hash
  $rv=$self->{RWL_dbh}->prepare("SELECT name FROM domains");
  $rv->execute;
  if($rv->rows > 0)
    {
    $self->{domains} = ();
    my $hash_ref;
    while($hash_ref = $rv->fetchrow_hashref)
      {
      $self->{domains}->{$hash_ref->{name}}='1';
      }
    }
  else
    {
    die "Oops! I can't build local domains hash.";
#    reconnect_to_db($self);
    }
  }


sub check_domain 
  {
  my ($self, $domainname) = @_;
  my $rv;
  my $result;
  if(!$domainname) {return 0;}
  $rv=$self->{RWL_dbh}->prepare("SELECT name FROM domains WHERE name=lower('$domainname')");
  $rv->execute;
  if($DBI::errstr)
    {
    reconnect_to_db($self);
#reconnect and try one more time
    $rv=$self->{RWL_dbh}->prepare("SELECT name FROM domains WHERE name=lower('$domainname')");
    $rv->execute;
    }
  $result = $rv->rows;
  $rv->finish() if($rv);
  if ($result > 0) {return 1;}
  return 0;
  }

sub hash_check_domain 
  {
  my ($self, $domainname) = @_;
  my $rv;
  my $result;
  if(!$domainname) {return 0;}
  if(exists($self->{domains}->{$domainname})) {return 1;}
  return 0;
  }

# and the eval rule itself
sub check_rwl_plugin {
  my ($self, $permsgstatus) = @_;
  my ($sec,$min,$hour,$mday,$mon,$year,$wday) = localtime(time);
  Mail::SpamAssassin::Logger::dbg ("RWL: eval test called: $self\n");
  # ... hard work goes here...
  connect_to_db($self) if (!$self->{RWL_dbh});

# call periodical tasks
  if($self->{Hour} != $hour)
    {
    cron_tasks($self);
    $self->{Hour} = $hour;
    }

  my $dbh = $self->{RWL_dbh};
  my $rv;
  my $flag = 0;
  my $mail_address;
  my $mail_address2;

  foreach $mail_address ($permsgstatus->all_from_addrs()) 
    {
  # $user."\@"
  # rotate through $domain and check
    my ($user,$domain) = split(/\@/, $mail_address,2);
    if (hash_check_domain($self, $domain))
      {
# Add/Update database
      Mail::SpamAssassin::Logger::info ("RWL: MAIL FROM's domain $mail_address is local.\n");
      foreach $mail_address2 ($permsgstatus->all_to_addrs())
	{
	my ($user,$domain) = split(/\@/, $mail_address2,2);
	if (hash_check_domain($self, $domain))
	  {
	  Mail::SpamAssassin::Logger::info ("RWL: MAIL RCPT: $mail_address2 is local. Ignore it.\n");
	  }
	else
	  {
################################################################################################
# PostgreSQL 
################################################################################################
	  $rv=$dbh->prepare("SELECT list_update('$mail_address2', 'now');");
	  $rv->execute;
#	  reconnect_to_db($self) if($DBI::errstr);
	  if($DBI::errstr)
	    {
	    reconnect_to_db($self);
#reconnect and try one more time
	    $rv->finish() if($rv);
	    $rv=$dbh->prepare("SELECT list_update('$mail_address2', 'now');");
	    $rv->execute;
	    }
	  $rv->finish() if($rv);
	  Mail::SpamAssassin::Logger::info ("RWL: MAIL RCPT: Add $mail_address2 to db.\n");

################################################################################################
# MySQL and other DB
################################################################################################
#          $rv=$dbh->prepare("SELECT email FROM list WHERE email=lower('$mail_address2')");
#          $rv->execute;
#          if ($rv->rows < 1)
#            {
#            $rv->finish() if($rv);
#            print "Mail::SpamAssassin::Plugin::RWL: MAIL RCPT: $mail_address2 not exist. Add them.\n";
#	     $rv=$dbh->prepare("INSERT INTO list (email, counter, first_detect, last_detect) VALUES (lower('$mail_address2'), 1, 'now', 'now')");
#            $rv->execute;
#            }
#          else
#            {
#            $rv->finish() if($rv);
#            print "Mail::SpamAssassin::Plugin::RWL: MAIL FROM: $mail_address2 is exist. Updating info\n";
#            $rv=$dbh->prepare("UPDATE list SET counter=counter+1, last_detect='now' WHERE email=lower('$mail_address2')");
#            $rv->execute;
#            }
#          $rv->finish() if($rv);
################################################################################################
	  }
	}
      }
    else
      {
# Read database
      Mail::SpamAssassin::Logger::info ("RWL: MAIL FORM's domain $mail_address is alien.\n");
      $rv=$dbh->prepare("SELECT email FROM list WHERE email=lower('$mail_address')");
      $rv->execute;
#	  reconnect_to_db($self) if($DBI::errstr);
      if($DBI::errstr)
	{
	reconnect_to_db($self);
#reconnect and try one more time
	$rv->finish() if($rv);
	$rv=$dbh->prepare("SELECT email FROM list WHERE email=lower('$mail_address')");
	$rv->execute;
	}
      if ($rv->rows > 0) 
        {
        Mail::SpamAssassin::Logger::info ("RWL: $mail_address - i know it.\n");
	$flag = 1;
	}
      $rv->finish() if($rv);
      }
    }
  return $flag;
}

1;

=back

=cut
